from __future__ import annotations

import json
import secrets
import uuid
from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from hashlib import sha256
from typing import Protocol

from app.core.auth.api_key_cache import get_api_key_cache
from app.core.cache.invalidation import NAMESPACE_API_KEY, get_cache_invalidation_poller
from app.core.usage.pricing import (
    UsageTokens,
    calculate_cost_from_usage,
    get_pricing_for_model,
)
from app.core.utils.time import to_utc_naive, utcnow
from app.db.models import Account, ApiKey, ApiKeyLimit, LimitType, LimitWindow
from app.modules.api_keys.repository import (
    _UNSET,
    ApiKeyTrendBucket,
    ApiKeyUsageSummary,
    ApiKeyUsageTotals,
    ReservationResult,
    UsageReservationData,
    UsageReservationItemData,
    _Unset,
)

_SPARKLINE_DAYS = 7
_DETAIL_BUCKET_SECONDS = 3600


class ApiKeysRepositoryProtocol(Protocol):
    async def create(self, row: ApiKey) -> ApiKey: ...

    async def get_by_id(self, key_id: str) -> ApiKey | None: ...

    async def get_by_hash(self, key_hash: str) -> ApiKey | None: ...

    async def list_all(self) -> list[ApiKey]: ...
    async def list_usage_summary_by_key(self) -> dict[str, ApiKeyUsageSummary]: ...
    async def get_usage_summary_by_key_id(self, key_id: str) -> ApiKeyUsageSummary: ...
    async def list_accounts_by_ids(self, account_ids: list[str]) -> list[Account]: ...

    async def update(
        self,
        key_id: str,
        *,
        name: str | _Unset = ...,
        allowed_models: str | None | _Unset = ...,
        enforced_model: str | None | _Unset = ...,
        enforced_reasoning_effort: str | None | _Unset = ...,
        enforced_service_tier: str | None | _Unset = ...,
        account_assignment_scope_enabled: bool | _Unset = ...,
        expires_at: datetime | None | _Unset = ...,
        is_active: bool | _Unset = ...,
        key_hash: str | _Unset = ...,
        key_prefix: str | _Unset = ...,
        commit: bool = True,
    ) -> ApiKey | None: ...

    async def delete(self, key_id: str) -> bool: ...

    async def update_last_used(self, key_id: str) -> None: ...

    async def commit(self) -> None: ...

    async def rollback(self) -> None: ...

    async def get_limits_by_key(self, key_id: str) -> list[ApiKeyLimit]: ...

    async def replace_limits(self, key_id: str, limits: list[ApiKeyLimit]) -> list[ApiKeyLimit]: ...

    async def upsert_limits(
        self, key_id: str, limits: list[ApiKeyLimit], *, commit: bool = True
    ) -> list[ApiKeyLimit]: ...
    async def replace_account_assignments(
        self, key_id: str, account_ids: list[str], *, commit: bool = True
    ) -> None: ...

    async def increment_limit_usage(
        self,
        key_id: str,
        *,
        model: str,
        input_tokens: int,
        output_tokens: int,
        cost_microdollars: int,
    ) -> None: ...

    async def reset_limit(self, limit_id: int, *, expected_reset_at: datetime, new_reset_at: datetime) -> bool: ...

    async def try_reserve_usage(
        self,
        limit_id: int,
        *,
        delta: int,
        expected_reset_at: datetime,
    ) -> ReservationResult: ...

    async def adjust_reserved_usage(
        self,
        limit_id: int,
        *,
        delta: int,
        expected_reset_at: datetime,
    ) -> bool: ...

    async def create_usage_reservation(
        self,
        reservation_id: str,
        *,
        key_id: str,
        model: str,
        items: list[UsageReservationItemData],
    ) -> None: ...

    async def get_usage_reservation(self, reservation_id: str) -> UsageReservationData | None: ...

    async def transition_usage_reservation_status(
        self,
        reservation_id: str,
        *,
        expected_status: str,
        new_status: str,
    ) -> bool: ...

    async def upsert_reservation_item_actual(
        self,
        reservation_id: str,
        *,
        item: UsageReservationItemData,
        actual_delta: int,
    ) -> None: ...

    async def settle_usage_reservation(
        self,
        reservation_id: str,
        *,
        status: str,
        input_tokens: int | None,
        output_tokens: int | None,
        cached_input_tokens: int | None,
        cost_microdollars: int | None,
    ) -> None: ...

    async def trends_by_key(
        self,
        key_id: str,
        since: datetime,
        until: datetime,
        bucket_seconds: int = 3600,
    ) -> list[ApiKeyTrendBucket]: ...

    async def usage_7d(
        self,
        key_id: str,
        since: datetime,
        until: datetime,
    ) -> ApiKeyUsageTotals: ...


class ApiKeyNotFoundError(ValueError):
    pass


class ApiKeyInvalidError(ValueError):
    pass


class ApiKeyRateLimitExceededError(ValueError):
    def __init__(self, *, message: str, reset_at: datetime) -> None:
        super().__init__(message)
        self.reset_at = reset_at


@dataclass(frozen=True, slots=True)
class LimitRuleData:
    id: int
    limit_type: str
    limit_window: str
    max_value: int
    current_value: int
    model_filter: str | None
    reset_at: datetime


@dataclass(frozen=True, slots=True)
class LimitRuleInput:
    limit_type: str
    limit_window: str
    max_value: int
    model_filter: str | None = None


@dataclass(frozen=True, slots=True)
class ApiKeyCreateData:
    name: str
    allowed_models: list[str] | None
    enforced_model: str | None = None
    enforced_reasoning_effort: str | None = None
    enforced_service_tier: str | None = None
    expires_at: datetime | None = None
    limits: list[LimitRuleInput] = field(default_factory=list)


@dataclass(frozen=True, slots=True)
class ApiKeyUpdateData:
    name: str | None = None
    name_set: bool = False
    allowed_models: list[str] | None = None
    allowed_models_set: bool = False
    enforced_model: str | None = None
    enforced_model_set: bool = False
    enforced_reasoning_effort: str | None = None
    enforced_reasoning_effort_set: bool = False
    enforced_service_tier: str | None = None
    enforced_service_tier_set: bool = False
    expires_at: datetime | None = None
    expires_at_set: bool = False
    is_active: bool | None = None
    is_active_set: bool = False
    assigned_account_ids: list[str] | None = None
    assigned_account_ids_set: bool = False
    limits: list[LimitRuleInput] | None = None
    limits_set: bool = False
    reset_usage: bool = False


@dataclass(frozen=True, slots=True)
class ApiKeyData:
    id: str
    name: str
    key_prefix: str
    allowed_models: list[str] | None
    enforced_model: str | None
    enforced_reasoning_effort: str | None
    enforced_service_tier: str | None
    expires_at: datetime | None
    is_active: bool
    created_at: datetime
    last_used_at: datetime | None
    limits: list[LimitRuleData] = field(default_factory=list)
    usage_summary: "ApiKeyUsageSummaryData | None" = None
    account_assignment_scope_enabled: bool = False
    assigned_account_ids: list[str] = field(default_factory=list)


@dataclass(frozen=True, slots=True)
class ApiKeyCreatedData(ApiKeyData):
    key: str = ""


@dataclass(frozen=True, slots=True)
class ApiKeyUsageSummaryData:
    request_count: int
    total_tokens: int
    cached_input_tokens: int
    total_cost_usd: float


@dataclass(frozen=True, slots=True)
class ApiKeyUsageReservationData:
    reservation_id: str
    key_id: str
    model: str


class ApiKeysService:
    def __init__(self, repository: ApiKeysRepositoryProtocol) -> None:
        self._repository = repository

    async def create_key(self, payload: ApiKeyCreateData) -> ApiKeyCreatedData:
        now = utcnow()
        expires_at = _normalize_expires_at(payload.expires_at)
        plain_key = _generate_plain_key()
        normalized_allowed_models = _normalize_allowed_models(payload.allowed_models)
        enforced_model = _normalize_model_slug(payload.enforced_model)
        enforced_reasoning_effort = _normalize_reasoning_effort(payload.enforced_reasoning_effort)
        enforced_service_tier = _normalize_service_tier(payload.enforced_service_tier)
        _validate_model_enforcement(enforced_model=enforced_model, allowed_models=normalized_allowed_models)
        row = ApiKey(
            id=str(__import__("uuid").uuid4()),
            name=_normalize_name(payload.name),
            key_hash=_hash_key(plain_key),
            key_prefix=plain_key[:15],
            allowed_models=_serialize_allowed_models(normalized_allowed_models),
            enforced_model=enforced_model,
            enforced_reasoning_effort=enforced_reasoning_effort,
            enforced_service_tier=enforced_service_tier,
            expires_at=expires_at,
            is_active=True,
            created_at=now,
            last_used_at=None,
        )
        created = await self._repository.create(row)

        if payload.limits:
            limit_rows = [_limit_input_to_row(li, created.id, now) for li in payload.limits]
            await self._repository.replace_limits(created.id, limit_rows)
            # Refresh to get updated limits
            created = await self._repository.get_by_id(created.id)
            if created is None:
                raise ValueError("Failed to create API key")

        return _to_created_data(_to_api_key_data(created), plain_key)

    async def list_keys(self) -> list[ApiKeyData]:
        rows = await self._repository.list_all()
        usage_summary_by_key = await self._repository.list_usage_summary_by_key()
        return [
            _to_api_key_data(row, usage_summary=_to_usage_summary_data(usage_summary_by_key.get(row.id)))
            for row in rows
        ]

    async def update_key(self, key_id: str, payload: ApiKeyUpdateData) -> ApiKeyData:
        expires_at = _normalize_expires_at(payload.expires_at) if payload.expires_at_set else None
        existing = await self._repository.get_by_id(key_id)
        if existing is None:
            raise ApiKeyNotFoundError(f"API key not found: {key_id}")

        if payload.allowed_models_set:
            allowed_models = _normalize_allowed_models(payload.allowed_models)
        else:
            allowed_models = None
        if payload.assigned_account_ids_set:
            assigned_account_ids = _normalize_assigned_account_ids(payload.assigned_account_ids)
            existing_accounts = await self._repository.list_accounts_by_ids(assigned_account_ids)
            existing_account_ids = {account.id for account in existing_accounts}
            missing_account_ids = [
                account_id for account_id in assigned_account_ids if account_id not in existing_account_ids
            ]
            if missing_account_ids:
                missing = ", ".join(missing_account_ids)
                raise ValueError(f"Unknown account ids: {missing}")
            account_assignment_scope_enabled: bool | _Unset = bool(assigned_account_ids)
        else:
            assigned_account_ids = None
            account_assignment_scope_enabled = _UNSET

        if payload.enforced_model_set:
            enforced_model = _normalize_model_slug(payload.enforced_model)
        else:
            enforced_model = None

        if payload.enforced_reasoning_effort_set:
            enforced_reasoning_effort = _normalize_reasoning_effort(payload.enforced_reasoning_effort)
        else:
            enforced_reasoning_effort = None

        if payload.enforced_service_tier_set:
            enforced_service_tier = _normalize_service_tier(payload.enforced_service_tier)
        else:
            enforced_service_tier = None

        if payload.allowed_models_set or payload.enforced_model_set:
            effective_allowed_models = (
                allowed_models if payload.allowed_models_set else _deserialize_allowed_models(existing.allowed_models)
            )
            effective_enforced_model = (
                enforced_model if payload.enforced_model_set else _normalize_model_slug(existing.enforced_model)
            )
            _validate_model_enforcement(
                enforced_model=effective_enforced_model,
                allowed_models=effective_allowed_models,
            )

        limit_rows: list[ApiKeyLimit] | None = None
        if payload.limits_set:
            now = utcnow()
            existing_limits = await self._repository.get_limits_by_key(key_id)
            submitted_limits = payload.limits or []
            limit_rows = _build_limit_rows_for_update(
                key_id=key_id,
                now=now,
                submitted_limits=submitted_limits,
                existing_limits=existing_limits,
                reset_usage=payload.reset_usage,
            )
        elif payload.reset_usage:
            now = utcnow()
            existing_limits = await self._repository.get_limits_by_key(key_id)
            limit_rows = _build_reset_limit_rows(key_id=key_id, now=now, existing_limits=existing_limits)

        try:
            row = await self._repository.update(
                key_id,
                name=_normalize_name(payload.name or "") if payload.name_set else _UNSET,
                allowed_models=_serialize_allowed_models(allowed_models) if payload.allowed_models_set else _UNSET,
                enforced_model=enforced_model if payload.enforced_model_set else _UNSET,
                enforced_reasoning_effort=(
                    enforced_reasoning_effort if payload.enforced_reasoning_effort_set else _UNSET
                ),
                enforced_service_tier=(enforced_service_tier if payload.enforced_service_tier_set else _UNSET),
                account_assignment_scope_enabled=account_assignment_scope_enabled,
                expires_at=expires_at if payload.expires_at_set else _UNSET,
                is_active=(payload.is_active if payload.is_active_set and payload.is_active is not None else _UNSET),
                commit=False,
            )
            if row is None:
                raise ApiKeyNotFoundError(f"API key not found: {key_id}")

            if payload.assigned_account_ids_set:
                assert assigned_account_ids is not None
                await self._repository.replace_account_assignments(key_id, assigned_account_ids, commit=False)

            if limit_rows is not None:
                await self._repository.upsert_limits(key_id, limit_rows, commit=False)

            await self._repository.commit()
        except Exception:
            await self._repository.rollback()
            raise

        if (
            payload.assigned_account_ids_set
            or limit_rows is not None
            or payload.name_set
            or payload.allowed_models_set
            or payload.enforced_model_set
            or payload.enforced_reasoning_effort_set
            or payload.enforced_service_tier_set
            or payload.expires_at_set
            or payload.is_active_set
        ):
            row = await self._repository.get_by_id(key_id)
            if row is None:
                raise ApiKeyNotFoundError(f"API key not found: {key_id}")

        await get_api_key_cache().invalidate(row.key_hash)
        poller = get_cache_invalidation_poller()
        if poller is not None:
            await poller.bump(NAMESPACE_API_KEY)
        return _to_api_key_data(row)

    async def delete_key(self, key_id: str) -> None:
        row = await self._repository.get_by_id(key_id)
        if row is None:
            raise ApiKeyNotFoundError(f"API key not found: {key_id}")
        deleted = await self._repository.delete(key_id)
        if not deleted:
            raise ApiKeyNotFoundError(f"API key not found: {key_id}")
        await get_api_key_cache().invalidate(row.key_hash)
        poller = get_cache_invalidation_poller()
        if poller is not None:
            await poller.bump(NAMESPACE_API_KEY)

    async def regenerate_key(self, key_id: str) -> ApiKeyCreatedData:
        row = await self._repository.get_by_id(key_id)
        if row is None:
            raise ApiKeyNotFoundError(f"API key not found: {key_id}")
        old_key_hash = row.key_hash
        plain_key = _generate_plain_key()
        updated = await self._repository.update(
            key_id,
            key_hash=_hash_key(plain_key),
            key_prefix=plain_key[:15],
        )
        if updated is None:
            raise ApiKeyNotFoundError(f"API key not found: {key_id}")
        await get_api_key_cache().invalidate(old_key_hash)
        poller = get_cache_invalidation_poller()
        if poller is not None:
            await poller.bump(NAMESPACE_API_KEY)
        return _to_created_data(_to_api_key_data(updated), plain_key)

    async def validate_key(self, plain_key: str) -> ApiKeyData:
        if not plain_key:
            raise ApiKeyInvalidError("Missing API key in Authorization header")

        key_hash = _hash_key(plain_key)
        now = utcnow()
        row = _ensure_valid_api_key_row(await self._repository.get_by_hash(key_hash))
        if row.expires_at is not None and row.expires_at < now:
            raise ApiKeyInvalidError("API key has expired")
        limits_reset = await _lazy_reset_expired_limits(self._repository, row.limits, now=now)
        refreshed = _ensure_valid_api_key_row(await self._repository.get_by_hash(key_hash)) if limits_reset else row
        if refreshed.expires_at is not None and refreshed.expires_at < now:
            raise ApiKeyInvalidError("API key has expired")
        return _to_api_key_data(refreshed)

    async def get_key_by_id(self, key_id: str) -> ApiKeyData:
        now = utcnow()
        row = _ensure_valid_api_key_row(await self._repository.get_by_id(key_id))
        if row.expires_at is not None and row.expires_at < now:
            raise ApiKeyInvalidError("API key has expired")
        return _to_api_key_data(row)

    async def enforce_limits_for_request(
        self,
        key_id: str,
        *,
        request_model: str | None,
        request_service_tier: str | None = None,
    ) -> ApiKeyUsageReservationData:
        now = utcnow()
        row = _ensure_valid_api_key_row(await self._repository.get_by_id(key_id))
        if row.expires_at is not None and row.expires_at < now:
            raise ApiKeyInvalidError("API key has expired")
        limits_reset = await _lazy_reset_expired_limits(self._repository, row.limits, now=now)
        refreshed = _ensure_valid_api_key_row(await self._repository.get_by_id(key_id)) if limits_reset else row
        if refreshed.expires_at is not None and refreshed.expires_at < now:
            raise ApiKeyInvalidError("API key has expired")

        reservation_items: list[UsageReservationItemData] = []
        try:
            for limit in refreshed.limits:
                if not _limit_applies_for_request(limit, request_model=request_model):
                    continue
                if limit.current_value >= limit.max_value:
                    raise _rate_limit_exceeded_error(limit)
                reserve_delta = _reserve_delta_for_limit(
                    limit,
                    request_model=request_model,
                    request_service_tier=request_service_tier,
                )
                if reserve_delta <= 0:
                    continue
                result = await self._repository.try_reserve_usage(
                    limit.id,
                    delta=reserve_delta,
                    expected_reset_at=limit.reset_at,
                )
                if not result.success:
                    raise _rate_limit_exceeded_error(limit)
                reservation_items.append(
                    UsageReservationItemData(
                        limit_id=limit.id,
                        limit_type=limit.limit_type,
                        reserved_delta=reserve_delta,
                        expected_reset_at=limit.reset_at,
                    )
                )

            reservation_id = _next_usage_reservation_id()
            await self._repository.create_usage_reservation(
                reservation_id,
                key_id=key_id,
                model=request_model or "",
                items=reservation_items,
            )
            await self._repository.commit()
        except Exception:
            await self._repository.rollback()
            raise

        return ApiKeyUsageReservationData(
            reservation_id=reservation_id,
            key_id=key_id,
            model=request_model or "",
        )

    async def finalize_usage_reservation(
        self,
        reservation_id: str,
        *,
        model: str,
        input_tokens: int,
        output_tokens: int,
        cached_input_tokens: int = 0,
        service_tier: str | None = None,
    ) -> None:
        await self._settle_usage_reservation(
            reservation_id,
            model=model,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            cached_input_tokens=cached_input_tokens,
            service_tier=service_tier,
            status="finalized",
        )

    async def fail_usage_reservation(
        self,
        reservation_id: str,
        *,
        model: str,
        input_tokens: int | None = None,
        output_tokens: int | None = None,
        cached_input_tokens: int | None = None,
        service_tier: str | None = None,
    ) -> None:
        await self._settle_usage_reservation(
            reservation_id,
            model=model,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            cached_input_tokens=cached_input_tokens,
            service_tier=service_tier,
            status="failed",
        )

    async def _settle_usage_reservation(
        self,
        reservation_id: str,
        *,
        model: str,
        input_tokens: int | None,
        output_tokens: int | None,
        cached_input_tokens: int | None,
        service_tier: str | None,
        status: str,
    ) -> None:
        reservation = await self._repository.get_usage_reservation(reservation_id)
        if reservation is None or reservation.status != "reserved":
            return

        claimed = await self._repository.transition_usage_reservation_status(
            reservation_id,
            expected_status="reserved",
            new_status="settling",
        )
        if not claimed:
            await self._repository.rollback()
            return

        effective_input_tokens = input_tokens or 0
        effective_output_tokens = output_tokens or 0
        effective_cached_input_tokens = cached_input_tokens or 0
        cost_microdollars = _calculate_cost_microdollars(
            model,
            effective_input_tokens,
            effective_output_tokens,
            effective_cached_input_tokens,
            service_tier,
        )

        try:
            for item in reservation.items:
                actual_delta = _compute_increment_for_limit_type(
                    item.limit_type,
                    input_tokens=effective_input_tokens,
                    output_tokens=effective_output_tokens,
                    cost_microdollars=cost_microdollars,
                )
                delta = actual_delta - item.reserved_delta
                if delta != 0:
                    await self._repository.adjust_reserved_usage(
                        item.limit_id,
                        delta=delta,
                        expected_reset_at=item.expected_reset_at,
                    )
                await self._repository.upsert_reservation_item_actual(
                    reservation_id,
                    item=item,
                    actual_delta=actual_delta,
                )

            await self._repository.settle_usage_reservation(
                reservation_id,
                status=status,
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                cached_input_tokens=cached_input_tokens,
                cost_microdollars=cost_microdollars,
            )
            await self._repository.commit()
        except Exception:
            await self._repository.rollback()
            raise

        await self._repository.update_last_used(reservation.api_key_id)

    async def release_usage_reservation(self, reservation_id: str) -> None:
        reservation = await self._repository.get_usage_reservation(reservation_id)
        if reservation is None or reservation.status != "reserved":
            return

        claimed = await self._repository.transition_usage_reservation_status(
            reservation_id,
            expected_status="reserved",
            new_status="released",
        )
        if not claimed:
            await self._repository.rollback()
            return

        try:
            for item in reservation.items:
                await self._repository.adjust_reserved_usage(
                    item.limit_id,
                    delta=-item.reserved_delta,
                    expected_reset_at=item.expected_reset_at,
                )
                await self._repository.upsert_reservation_item_actual(
                    reservation_id,
                    item=item,
                    actual_delta=0,
                )
            await self._repository.settle_usage_reservation(
                reservation_id,
                status="released",
                input_tokens=None,
                output_tokens=None,
                cached_input_tokens=None,
                cost_microdollars=None,
            )
            await self._repository.commit()
        except Exception:
            await self._repository.rollback()
            raise

    async def record_usage(
        self,
        key_id: str,
        *,
        model: str,
        input_tokens: int,
        output_tokens: int,
        cached_input_tokens: int = 0,
        service_tier: str | None = None,
    ) -> None:
        cost_microdollars = _calculate_cost_microdollars(
            model,
            input_tokens,
            output_tokens,
            cached_input_tokens,
            service_tier,
        )
        await self._repository.increment_limit_usage(
            key_id,
            model=model,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            cost_microdollars=cost_microdollars,
        )

    async def get_key_trends(self, key_id: str) -> ApiKeyTrendsData | None:
        row = await self._repository.get_by_id(key_id)
        if row is None:
            return None
        now = utcnow()
        since = now - timedelta(days=_SPARKLINE_DAYS)
        buckets = await self._repository.trends_by_key(
            key_id,
            since,
            now,
            _DETAIL_BUCKET_SECONDS,
        )
        return _build_api_key_trends(key_id, buckets, since, now, _DETAIL_BUCKET_SECONDS)

    async def get_key_usage_summary_for_self(self, key_id: str) -> ApiKeySelfUsageData | None:
        """Return usage summary + current limits for a single key (self-service lookup)."""
        row = await self._repository.get_by_id(key_id)
        if row is None:
            return None

        now = utcnow()
        # Reset any expired limits before reading state
        limits_reset = await _lazy_reset_expired_limits(self._repository, row.limits, now=now)
        refreshed = await self._repository.get_by_id(key_id) if limits_reset else row
        if refreshed is None:
            return None

        usage = await self._repository.get_usage_summary_by_key_id(key_id)
        limits = [
            ApiKeySelfLimitData(
                limit_type=limit.limit_type.value,
                limit_window=limit.limit_window.value,
                max_value=limit.max_value,
                current_value=max(0, min(limit.current_value, limit.max_value)),
                remaining_value=max(0, limit.max_value - max(0, min(limit.current_value, limit.max_value))),
                model_filter=limit.model_filter,
                reset_at=limit.reset_at,
                source="api_key_override" if limit.limit_type == LimitType.CREDITS else "api_key_limit",
            )
            for limit in refreshed.limits
        ]
        return ApiKeySelfUsageData(
            request_count=usage.request_count,
            total_tokens=usage.total_tokens,
            cached_input_tokens=usage.cached_input_tokens,
            total_cost_usd=usage.total_cost_usd,
            limits=limits,
        )

    async def get_key_usage_7d(self, key_id: str) -> ApiKeyUsage7DayData | None:
        row = await self._repository.get_by_id(key_id)
        if row is None:
            return None
        now = utcnow()
        since = now - timedelta(days=7)
        data = await self._repository.usage_7d(key_id, since, now)
        return ApiKeyUsage7DayData(
            key_id=key_id,
            total_tokens=data.total_tokens,
            total_cost_usd=data.total_cost_usd,
            total_requests=data.total_requests,
            cached_input_tokens=data.cached_input_tokens,
        )


@dataclass(frozen=True, slots=True)
class ApiKeyTrendsPoint:
    t: datetime
    v: float


@dataclass(frozen=True, slots=True)
class ApiKeyTrendsData:
    key_id: str
    cost: list[ApiKeyTrendsPoint] = field(default_factory=list)
    tokens: list[ApiKeyTrendsPoint] = field(default_factory=list)


@dataclass(frozen=True, slots=True)
class ApiKeyUsage7DayData:
    key_id: str
    total_tokens: int = 0
    total_cost_usd: float = 0.0
    total_requests: int = 0
    cached_input_tokens: int = 0


@dataclass(frozen=True, slots=True)
class ApiKeySelfLimitData:
    limit_type: str
    limit_window: str
    max_value: int
    current_value: int
    remaining_value: int
    model_filter: str | None
    reset_at: datetime
    source: str = "api_key_limit"


@dataclass(frozen=True, slots=True)
class ApiKeySelfUsageData:
    request_count: int = 0
    total_tokens: int = 0
    cached_input_tokens: int = 0
    total_cost_usd: float = 0.0
    limits: list[ApiKeySelfLimitData] = field(default_factory=list)


def _normalize_name(name: str) -> str:
    normalized = name.strip()
    if not normalized:
        raise ValueError("API key name is required")
    return normalized


def _generate_plain_key() -> str:
    return f"sk-clb-{secrets.token_urlsafe(32)}"


def _hash_key(plain_key: str) -> str:
    return sha256(plain_key.encode("utf-8")).hexdigest()


def _serialize_allowed_models(allowed_models: list[str] | None) -> str | None:
    if allowed_models is None:
        return None
    return json.dumps(allowed_models)


def _deserialize_allowed_models(payload: str | None) -> list[str] | None:
    if payload is None:
        return None
    parsed = json.loads(payload)
    if not isinstance(parsed, list):
        return None
    models = [str(value).strip() for value in parsed if str(value).strip()]
    return models


def _normalize_allowed_models(allowed_models: list[str] | None) -> list[str] | None:
    if allowed_models is None:
        return None
    return [model.strip() for model in allowed_models if model and model.strip()]


def _normalize_assigned_account_ids(account_ids: list[str] | None) -> list[str]:
    if not account_ids:
        return []
    normalized: list[str] = []
    seen: set[str] = set()
    for account_id in account_ids:
        value = account_id.strip()
        if not value or value in seen:
            continue
        normalized.append(value)
        seen.add(value)
    return normalized


def _normalize_model_slug(value: str | None) -> str | None:
    if value is None:
        return None
    normalized = value.strip()
    if not normalized:
        return None
    return normalized


_SUPPORTED_REASONING_EFFORTS = frozenset({"none", "minimal", "low", "medium", "high", "xhigh"})
_SUPPORTED_SERVICE_TIERS = frozenset({"auto", "default", "priority", "flex"})


def _normalize_expires_at(value: datetime | None) -> datetime | None:
    if value is None:
        return None
    return to_utc_naive(value)


def _normalize_reasoning_effort(value: str | None) -> str | None:
    if value is None:
        return None
    normalized = value.strip().lower()
    if not normalized:
        return None
    if normalized not in _SUPPORTED_REASONING_EFFORTS:
        options = ", ".join(sorted(_SUPPORTED_REASONING_EFFORTS))
        raise ValueError(f"Unsupported enforced reasoning effort '{normalized}'. Expected one of: {options}")
    return normalized


def _normalize_reasoning_effort_lenient(value: str | None) -> str | None:
    if value is None:
        return None
    normalized = value.strip().lower()
    if not normalized:
        return None
    if normalized in _SUPPORTED_REASONING_EFFORTS:
        return normalized
    return None


def _normalize_service_tier(value: str | None) -> str | None:
    if value is None:
        return None
    normalized = value.strip().lower()
    if not normalized:
        return None
    if normalized == "fast":
        normalized = "priority"
    if normalized not in _SUPPORTED_SERVICE_TIERS:
        options = ", ".join(sorted(_SUPPORTED_SERVICE_TIERS | {"fast"}))
        raise ValueError(f"Unsupported enforced service tier '{normalized}'. Expected one of: {options}")
    return normalized


def _normalize_service_tier_lenient(value: str | None) -> str | None:
    if value is None:
        return None
    normalized = value.strip().lower()
    if not normalized:
        return None
    if normalized == "fast":
        return "priority"
    if normalized in _SUPPORTED_SERVICE_TIERS:
        return normalized
    return None


def _validate_model_enforcement(*, enforced_model: str | None, allowed_models: list[str] | None) -> None:
    if enforced_model is None or not allowed_models:
        return
    if enforced_model not in allowed_models:
        raise ValueError("enforced_model must be present in allowed_models when allowed_models is configured")


def _to_limit_rule_data(limit: ApiKeyLimit) -> LimitRuleData:
    return LimitRuleData(
        id=limit.id,
        limit_type=limit.limit_type.value,
        limit_window=limit.limit_window.value,
        max_value=limit.max_value,
        current_value=limit.current_value,
        model_filter=limit.model_filter,
        reset_at=limit.reset_at,
    )


def _ensure_valid_api_key_row(row: ApiKey | None) -> ApiKey:
    if row is None or not row.is_active:
        raise ApiKeyInvalidError("Invalid API key")
    return row


async def _lazy_reset_expired_limits(
    repository: ApiKeysRepositoryProtocol,
    limits: list[ApiKeyLimit],
    *,
    now: datetime,
) -> bool:
    reset_performed = False
    for limit in limits:
        if limit.reset_at >= now:
            continue
        new_reset_at = _advance_reset(limit.reset_at, now, limit.limit_window)
        await repository.reset_limit(
            limit.id,
            expected_reset_at=limit.reset_at,
            new_reset_at=new_reset_at,
        )
        reset_performed = True
    return reset_performed


def _rate_limit_exceeded_error(limit: ApiKeyLimit) -> ApiKeyRateLimitExceededError:
    return ApiKeyRateLimitExceededError(
        message=f"API key {limit.limit_type.value} {limit.limit_window.value} limit exceeded"
        + (f" for model {limit.model_filter}" if limit.model_filter else ""),
        reset_at=limit.reset_at,
    )


def _limit_applies_for_request(limit: ApiKeyLimit, *, request_model: str | None) -> bool:
    if limit.model_filter is None:
        return True
    if request_model is None:
        return False
    return limit.model_filter == request_model


def _reserve_delta_for_limit(
    limit: ApiKeyLimit,
    *,
    request_model: str | None,
    request_service_tier: str | None,
) -> int:
    remaining = limit.max_value - limit.current_value
    if remaining <= 0:
        return 0
    budget = _reserve_budget_for_limit_type(
        limit.limit_type,
        request_model=request_model,
        request_service_tier=request_service_tier,
    )
    return min(remaining, budget)


def _reserve_budget_for_limit_type(
    limit_type: LimitType,
    *,
    request_model: str | None,
    request_service_tier: str | None,
) -> int:
    if limit_type == LimitType.TOTAL_TOKENS:
        return 8_192
    if limit_type == LimitType.INPUT_TOKENS:
        return 8_192
    if limit_type == LimitType.OUTPUT_TOKENS:
        return 8_192
    if limit_type == LimitType.COST_USD:
        return _reserve_cost_budget_microdollars(request_model, request_service_tier)
    if limit_type == LimitType.CREDITS:
        return 0
    return 1


def _reserve_cost_budget_microdollars(model: str | None, service_tier: str | None) -> int:
    if not model:
        return 2_000_000
    cost_microdollars = _calculate_cost_microdollars(
        model,
        8_192,
        8_192,
        0,
        service_tier,
    )
    return cost_microdollars if cost_microdollars > 0 else 2_000_000


def _compute_increment_for_limit_type(
    limit_type: LimitType,
    *,
    input_tokens: int,
    output_tokens: int,
    cost_microdollars: int,
) -> int:
    if limit_type == LimitType.TOTAL_TOKENS:
        return input_tokens + output_tokens
    if limit_type == LimitType.INPUT_TOKENS:
        return input_tokens
    if limit_type == LimitType.OUTPUT_TOKENS:
        return output_tokens
    if limit_type == LimitType.COST_USD:
        return cost_microdollars
    if limit_type == LimitType.CREDITS:
        return 0
    return 0


def _next_usage_reservation_id() -> str:
    return f"ur_{uuid.uuid4().hex}"


def _to_created_data(data: ApiKeyData, key: str) -> ApiKeyCreatedData:
    return ApiKeyCreatedData(
        id=data.id,
        name=data.name,
        key_prefix=data.key_prefix,
        allowed_models=data.allowed_models,
        enforced_model=data.enforced_model,
        enforced_reasoning_effort=data.enforced_reasoning_effort,
        enforced_service_tier=data.enforced_service_tier,
        expires_at=data.expires_at,
        is_active=data.is_active,
        created_at=data.created_at,
        last_used_at=data.last_used_at,
        limits=data.limits,
        usage_summary=data.usage_summary,
        account_assignment_scope_enabled=data.account_assignment_scope_enabled,
        assigned_account_ids=data.assigned_account_ids,
        key=key,
    )


def _to_api_key_data(row: ApiKey, *, usage_summary: ApiKeyUsageSummaryData | None = None) -> ApiKeyData:
    limits = [_to_limit_rule_data(limit) for limit in row.limits] if row.limits else []
    account_assignments = getattr(row, "account_assignments", [])
    return ApiKeyData(
        id=row.id,
        name=row.name,
        key_prefix=row.key_prefix,
        allowed_models=_deserialize_allowed_models(row.allowed_models),
        enforced_model=_normalize_model_slug(row.enforced_model),
        enforced_reasoning_effort=_normalize_reasoning_effort_lenient(row.enforced_reasoning_effort),
        enforced_service_tier=_normalize_service_tier_lenient(row.enforced_service_tier),
        expires_at=row.expires_at,
        is_active=row.is_active,
        created_at=row.created_at,
        last_used_at=row.last_used_at,
        limits=limits,
        usage_summary=usage_summary,
        account_assignment_scope_enabled=getattr(row, "account_assignment_scope_enabled", False),
        assigned_account_ids=[assignment.account_id for assignment in account_assignments],
    )


def _to_usage_summary_data(summary: ApiKeyUsageSummary | None) -> ApiKeyUsageSummaryData | None:
    if summary is None:
        return None
    return ApiKeyUsageSummaryData(
        request_count=summary.request_count,
        total_tokens=summary.total_tokens,
        cached_input_tokens=summary.cached_input_tokens,
        total_cost_usd=summary.total_cost_usd,
    )


def _limit_input_to_row(
    li: LimitRuleInput,
    key_id: str,
    now: datetime,
    *,
    current_value: int = 0,
    reset_at: datetime | None = None,
) -> ApiKeyLimit:
    window = LimitWindow(li.limit_window)
    if li.limit_type == LimitType.CREDITS.value and li.model_filter is not None:
        raise ValueError("credits limits do not support model_filter")
    return ApiKeyLimit(
        api_key_id=key_id,
        limit_type=LimitType(li.limit_type),
        limit_window=window,
        max_value=li.max_value,
        current_value=current_value,
        model_filter=li.model_filter,
        reset_at=reset_at if reset_at is not None else _next_reset(now, window),
    )


def _build_limit_rows_for_update(
    *,
    key_id: str,
    now: datetime,
    submitted_limits: list[LimitRuleInput],
    existing_limits: list[ApiKeyLimit],
    reset_usage: bool,
) -> list[ApiKeyLimit]:
    existing_by_key = {_limit_identity_from_row(limit): limit for limit in existing_limits}
    submitted_by_key = {_limit_identity_from_input(limit): limit for limit in submitted_limits}
    if len(submitted_by_key) != len(submitted_limits):
        raise ValueError("Duplicate limit rules are not allowed")

    rows: list[ApiKeyLimit] = []
    for submitted in submitted_limits:
        identity = _limit_identity_from_input(submitted)
        matched = existing_by_key.get(identity)
        if matched is None or reset_usage:
            rows.append(_limit_input_to_row(submitted, key_id, now))
            continue
        rows.append(
            _limit_input_to_row(
                submitted,
                key_id,
                now,
                current_value=matched.current_value,
                reset_at=matched.reset_at,
            )
        )
    return rows


def _build_reset_limit_rows(
    *,
    key_id: str,
    now: datetime,
    existing_limits: list[ApiKeyLimit],
) -> list[ApiKeyLimit]:
    rows: list[ApiKeyLimit] = []
    for existing in existing_limits:
        rows.append(
            ApiKeyLimit(
                api_key_id=key_id,
                limit_type=existing.limit_type,
                limit_window=existing.limit_window,
                max_value=existing.max_value,
                current_value=0,
                model_filter=existing.model_filter,
                reset_at=_next_reset(now, existing.limit_window),
            )
        )
    return rows


def _limit_identity_from_input(limit: LimitRuleInput) -> tuple[str, str, str | None]:
    return (limit.limit_type, limit.limit_window, limit.model_filter)


def _limit_identity_from_row(limit: ApiKeyLimit) -> tuple[str, str, str | None]:
    return (limit.limit_type.value, limit.limit_window.value, limit.model_filter)


def _next_reset(now: datetime, window: LimitWindow) -> datetime:
    if window == LimitWindow.FIVE_HOURS:
        return now + timedelta(hours=5)
    if window == LimitWindow.SEVEN_DAYS:
        return now + timedelta(days=7)
    if window == LimitWindow.DAILY:
        return now + timedelta(days=1)
    if window == LimitWindow.WEEKLY:
        return now + timedelta(days=7)
    if window == LimitWindow.MONTHLY:
        return now + timedelta(days=30)
    return now + timedelta(days=7)


def _advance_reset(reset_at: datetime, now: datetime, window: LimitWindow) -> datetime:
    delta = _window_delta(window)
    next_reset = reset_at
    while next_reset <= now:
        next_reset += delta
    return next_reset


def _window_delta(window: LimitWindow) -> timedelta:
    if window == LimitWindow.FIVE_HOURS:
        return timedelta(hours=5)
    if window == LimitWindow.SEVEN_DAYS:
        return timedelta(days=7)
    if window == LimitWindow.DAILY:
        return timedelta(days=1)
    if window == LimitWindow.WEEKLY:
        return timedelta(days=7)
    if window == LimitWindow.MONTHLY:
        return timedelta(days=30)
    return timedelta(days=7)


def _calculate_cost_microdollars(
    model: str,
    input_tokens: int,
    output_tokens: int,
    cached_input_tokens: int,
    service_tier: str | None = None,
) -> int:
    resolved = get_pricing_for_model(model)
    if resolved is None:
        return 0
    _, price = resolved
    usage = UsageTokens(
        input_tokens=float(input_tokens),
        output_tokens=float(output_tokens),
        cached_input_tokens=float(cached_input_tokens),
    )
    cost_usd = calculate_cost_from_usage(usage, price, service_tier=service_tier)
    if cost_usd is None:
        return 0
    return int(cost_usd * 1_000_000)


def _build_api_key_trends(
    key_id: str,
    buckets: list[ApiKeyTrendBucket],
    since: datetime,
    until: datetime,
    bucket_seconds: int,
) -> ApiKeyTrendsData:
    since_utc = since.replace(tzinfo=timezone.utc) if since.tzinfo is None else since.astimezone(timezone.utc)
    until_utc = until.replace(tzinfo=timezone.utc) if until.tzinfo is None else until.astimezone(timezone.utc)
    if until_utc <= since_utc:
        return ApiKeyTrendsData(key_id=key_id)

    start_epoch = (int(since_utc.timestamp()) // bucket_seconds) * bucket_seconds
    # The SQL window is exclusive of `until`, so step back one microsecond to find the last visible bucket.
    end_epoch = (int((until_utc - timedelta(microseconds=1)).timestamp()) // bucket_seconds) * bucket_seconds
    bucket_count = ((end_epoch - start_epoch) // bucket_seconds) + 1
    time_grid = [start_epoch + i * bucket_seconds for i in range(bucket_count)]

    cost_by_bucket: dict[int, float] = {}
    tokens_by_bucket: dict[int, int] = {}
    for b in buckets:
        cost_by_bucket[b.bucket_epoch] = b.total_cost_usd
        tokens_by_bucket[b.bucket_epoch] = b.total_tokens

    cost_points: list[ApiKeyTrendsPoint] = []
    tokens_points: list[ApiKeyTrendsPoint] = []
    for epoch in time_grid:
        dt = datetime.fromtimestamp(epoch, tz=timezone.utc)
        cost_points.append(ApiKeyTrendsPoint(t=dt, v=round(cost_by_bucket.get(epoch, 0.0), 6)))
        tokens_points.append(ApiKeyTrendsPoint(t=dt, v=float(tokens_by_bucket.get(epoch, 0))))

    return ApiKeyTrendsData(key_id=key_id, cost=cost_points, tokens=tokens_points)



from __future__ import annotations

import asyncio
from unittest.mock import MagicMock

import aiohttp
import pytest

from app.core.resilience import circuit_breaker as cb_module
from app.core.resilience.circuit_breaker import CircuitBreaker, CircuitBreakerOpenError, CircuitState

pytestmark = pytest.mark.unit


@pytest.mark.asyncio
async def test_circuit_opens_after_failure_threshold() -> None:
    breaker = CircuitBreaker(failure_threshold=5, recovery_timeout_seconds=60)

    async def fail() -> None:
        raise asyncio.TimeoutError

    for _ in range(5):
        with pytest.raises(asyncio.TimeoutError):
            await breaker.call(fail())

    assert breaker.state == CircuitState.OPEN


@pytest.mark.asyncio
async def test_open_transitions_to_half_open_after_timeout(monkeypatch: pytest.MonkeyPatch) -> None:
    breaker = CircuitBreaker(failure_threshold=5, recovery_timeout_seconds=10)
    breaker._state = CircuitState.OPEN
    breaker._last_failure_time = 0.0

    monkeypatch.setattr(cb_module.time, "monotonic", lambda: 11.0)

    assert breaker.state == CircuitState.HALF_OPEN


@pytest.mark.asyncio
async def test_half_open_closes_after_success_threshold(monkeypatch: pytest.MonkeyPatch) -> None:
    breaker = CircuitBreaker(failure_threshold=5, recovery_timeout_seconds=10, success_threshold=2)
    breaker._state = CircuitState.OPEN
    breaker._failure_count = 5
    breaker._last_failure_time = 0.0

    monkeypatch.setattr(cb_module.time, "monotonic", lambda: 11.0)

    async def succeed() -> str:
        return "ok"

    assert await breaker.call(succeed()) == "ok"
    assert breaker.state == CircuitState.HALF_OPEN
    assert await breaker.call(succeed()) == "ok"
    assert breaker.state == CircuitState.CLOSED


@pytest.mark.asyncio
async def test_4xx_errors_do_not_count_as_failures() -> None:
    breaker = CircuitBreaker(failure_threshold=5, recovery_timeout_seconds=60)

    request_info = MagicMock()
    request_info.real_url = "https://example.test"

    async def fail_4xx() -> None:
        raise aiohttp.ClientResponseError(
            request_info=request_info,
            history=(),
            status=404,
            message="not found",
            headers=None,
        )

    for _ in range(5):
        with pytest.raises(aiohttp.ClientResponseError):
            await breaker.call(fail_4xx())

    assert breaker.state == CircuitState.CLOSED


@pytest.mark.asyncio
async def test_open_state_fast_fails_requests(monkeypatch: pytest.MonkeyPatch) -> None:
    breaker = CircuitBreaker(failure_threshold=5, recovery_timeout_seconds=60)
    breaker._state = CircuitState.OPEN
    breaker._last_failure_time = 0.0
    monkeypatch.setattr(cb_module.time, "monotonic", lambda: 1.0)

    class ShouldNotBeAwaited:
        def __await__(self):
            raise AssertionError("request coroutine must not run when circuit is open")

    with pytest.raises(CircuitBreakerOpenError):
        await breaker.call(ShouldNotBeAwaited())



from __future__ import annotations

from datetime import datetime

from fastapi import APIRouter, Depends, Query

from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.dependencies import RequestLogsContext, get_request_logs_context
from app.modules.request_logs.schemas import (
    RequestLogFilterOptionsResponse,
    RequestLogModelOption,
    RequestLogsResponse,
)
from app.modules.request_logs.service import RequestLogModelOption as ServiceRequestLogModelOption

router = APIRouter(
    prefix="/api/request-logs",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)

_MODEL_OPTION_DELIMITER = ":::"


def _parse_model_option(value: str) -> ServiceRequestLogModelOption | None:
    raw = (value or "").strip()
    if not raw:
        return None
    if _MODEL_OPTION_DELIMITER not in raw:
        return ServiceRequestLogModelOption(model=raw, reasoning_effort=None)
    model, effort = raw.split(_MODEL_OPTION_DELIMITER, 1)
    model = model.strip()
    effort = effort.strip()
    if not model:
        return None
    return ServiceRequestLogModelOption(model=model, reasoning_effort=effort or None)


@router.get("", response_model=RequestLogsResponse)
async def list_request_logs(
    limit: int = Query(50, ge=1, le=1000),
    offset: int = Query(0, ge=0),
    search: str | None = Query(default=None),
    account_id: list[str] | None = Query(default=None, alias="accountId"),
    status: list[str] | None = Query(default=None),
    model: list[str] | None = Query(default=None),
    reasoning_effort: list[str] | None = Query(default=None, alias="reasoningEffort"),
    model_option: list[str] | None = Query(default=None, alias="modelOption"),
    since: datetime | None = Query(default=None),
    until: datetime | None = Query(default=None),
    context: RequestLogsContext = Depends(get_request_logs_context),
) -> RequestLogsResponse:
    parsed_options: list[ServiceRequestLogModelOption] | None = None
    if model_option:
        parsed = [_parse_model_option(value) for value in model_option]
        parsed_options = [value for value in parsed if value is not None] or None
    page = await context.service.list_recent(
        limit=limit,
        offset=offset,
        search=search,
        since=since,
        until=until,
        account_ids=account_id,
        model_options=parsed_options,
        models=model,
        reasoning_efforts=reasoning_effort,
        status=status,
    )
    return RequestLogsResponse(
        requests=page.requests,
        total=page.total,
        has_more=page.has_more,
    )


@router.get("/options", response_model=RequestLogFilterOptionsResponse)
async def list_request_log_filter_options(
    status: list[str] | None = Query(default=None),
    account_id: list[str] | None = Query(default=None, alias="accountId"),
    model: list[str] | None = Query(default=None),
    reasoning_effort: list[str] | None = Query(default=None, alias="reasoningEffort"),
    model_option: list[str] | None = Query(default=None, alias="modelOption"),
    since: datetime | None = Query(default=None),
    until: datetime | None = Query(default=None),
    context: RequestLogsContext = Depends(get_request_logs_context),
) -> RequestLogFilterOptionsResponse:
    _ = status  # Keep input backward compatible but do not self-filter status facet.
    parsed_options: list[ServiceRequestLogModelOption] | None = None
    if model_option:
        parsed = [_parse_model_option(value) for value in model_option]
        parsed_options = [value for value in parsed if value is not None] or None
    options = await context.service.list_filter_options(
        since=since,
        until=until,
        account_ids=account_id,
        model_options=parsed_options,
        models=model,
        reasoning_efforts=reasoning_effort,
    )
    return RequestLogFilterOptionsResponse(
        account_ids=options.account_ids,
        model_options=[
            RequestLogModelOption(model=option.model, reasoning_effort=option.reasoning_effort)
            for option in options.model_options
        ],
        statuses=options.statuses,
    )



from __future__ import annotations

from fastapi import APIRouter, Depends, Query

from app.core.auth.dependencies import set_dashboard_error_format, validate_dashboard_session
from app.dependencies import UsageContext, get_usage_context
from app.modules.usage.schemas import UsageHistoryResponse, UsageSummaryResponse, UsageWindowResponse

router = APIRouter(
    prefix="/api/usage",
    tags=["dashboard"],
    dependencies=[Depends(validate_dashboard_session), Depends(set_dashboard_error_format)],
)


@router.get("/summary", response_model=UsageSummaryResponse)
async def get_usage_summary(
    context: UsageContext = Depends(get_usage_context),
) -> UsageSummaryResponse:
    return await context.service.get_usage_summary()


@router.get("/history", response_model=UsageHistoryResponse)
async def get_usage_history(
    hours: int = Query(24, ge=1, le=168),
    context: UsageContext = Depends(get_usage_context),
) -> UsageHistoryResponse:
    return await context.service.get_usage_history(hours)


@router.get("/window", response_model=UsageWindowResponse)
async def get_usage_window(
    window: str = Query("primary", pattern="^(primary|secondary)$"),
    context: UsageContext = Depends(get_usage_context),
) -> UsageWindowResponse:
    return await context.service.get_usage_window(window)



from __future__ import annotations

import asyncio

from starlette.types import ASGIApp, Receive, Scope, Send


async def _reject_websocket(receive: Receive, send: Send, *, reason: str) -> None:
    try:
        event = await receive()
        if event.get("type") != "websocket.connect":
            return
    except Exception:
        return
    await send({"type": "websocket.close", "code": 1013, "reason": reason})


class BackpressureMiddleware:
    def __init__(self, app: ASGIApp, *, max_concurrent: int) -> None:
        self.app = app
        self._semaphore = asyncio.Semaphore(max_concurrent)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        path = scope.get("path", "")
        if path.startswith("/health"):
            await self.app(scope, receive, send)
            return

        if self._semaphore._value <= 0:
            if scope["type"] == "websocket":
                await _reject_websocket(receive, send, reason="Too Many Requests")
                return
            await send(
                {
                    "type": "http.response.start",
                    "status": 429,
                    "headers": [(b"retry-after", b"5"), (b"content-type", b"application/json")],
                }
            )
            await send({"type": "http.response.body", "body": b'{"detail":"Too Many Requests"}'})
            return

        await self._semaphore.acquire()
        try:
            await self.app(scope, receive, send)
        finally:
            self._semaphore.release()


__all__ = ["BackpressureMiddleware"]



from __future__ import annotations

from collections.abc import AsyncIterator
from datetime import UTC, datetime, timedelta
from pathlib import Path

import pytest
import sqlalchemy as sa
from alembic import command
from alembic.config import Config
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine

from app.core.exceptions import DashboardRateLimitError
from app.core.rate_limiter.db_rate_limiter import DatabaseRateLimiter
from app.db.models import Base, RateLimitAttempt

pytestmark = pytest.mark.unit


@pytest.fixture
async def async_session_factory() -> AsyncIterator[async_sessionmaker[AsyncSession]]:
    engine = create_async_engine("sqlite+aiosqlite:///:memory:")
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    factory = async_sessionmaker(engine, expire_on_commit=False)
    try:
        yield factory
    finally:
        await engine.dispose()


@pytest.mark.asyncio
async def test_single_instance_blocks_after_eight_attempts(
    async_session_factory: async_sessionmaker[AsyncSession],
) -> None:
    limiter = DatabaseRateLimiter(max_attempts=8, window_seconds=60, type="totp")

    async with async_session_factory() as session:
        for _ in range(8):
            await limiter.check_and_record("ip:single", session)

        with pytest.raises(DashboardRateLimitError):
            await limiter.check_and_record("ip:single", session)


@pytest.mark.asyncio
async def test_cross_replica_combined_attempts_are_enforced(
    async_session_factory: async_sessionmaker[AsyncSession],
) -> None:
    replica_one = DatabaseRateLimiter(max_attempts=8, window_seconds=60, type="password")
    replica_two = DatabaseRateLimiter(max_attempts=8, window_seconds=60, type="password")

    async with async_session_factory() as session_one:
        for _ in range(4):
            await replica_one.check_and_record("ip:replica", session_one)

    async with async_session_factory() as session_two:
        for _ in range(4):
            await replica_two.check_and_record("ip:replica", session_two)

    async with async_session_factory() as session_one_again:
        with pytest.raises(DashboardRateLimitError):
            await replica_one.check_and_record("ip:replica", session_one_again)


@pytest.mark.asyncio
async def test_window_expiry_ignores_old_attempts(
    async_session_factory: async_sessionmaker[AsyncSession],
) -> None:
    limiter = DatabaseRateLimiter(max_attempts=1, window_seconds=60, type="totp")

    async with async_session_factory() as session:
        old_attempt = RateLimitAttempt(
            key="ip:expired",
            type="totp",
            attempted_at=datetime.now(UTC) - timedelta(minutes=2),
        )
        session.add(old_attempt)
        await session.commit()

        await limiter.check_and_record("ip:expired", session)
        with pytest.raises(DashboardRateLimitError):
            await limiter.check_and_record("ip:expired", session)


def test_migration_upgrade_downgrade_upgrade_is_reversible(tmp_path: Path) -> None:
    db_path = tmp_path / "rate_limit_migration.db"
    cfg = Config()
    cfg.set_main_option("script_location", str((Path(__file__).resolve().parents[2] / "app/db/alembic").resolve()))
    cfg.set_main_option("sqlalchemy.url", f"sqlite:///{db_path}")

    command.upgrade(cfg, "head")
    command.downgrade(cfg, "base")
    command.upgrade(cfg, "head")

    engine = sa.create_engine(f"sqlite:///{db_path}", future=True)
    try:
        inspector = sa.inspect(engine)
        assert inspector.has_table("rate_limit_attempts") is True
    finally:
        engine.dispose()


# RED (xfail) tests proving the bug: successful logins are counted toward lockout


@pytest.mark.asyncio
async def test_successful_login_not_counted_toward_lockout(
    async_session_factory: async_sessionmaker[AsyncSession],
) -> None:
    """8 successful logins must NOT cause lockout; only failures count."""
    limiter = DatabaseRateLimiter(max_attempts=8, window_seconds=60, type="password")

    async with async_session_factory() as session:
        # Simulate 8 successful logins: check (no record) then clear on success
        for _ in range(8):
            await limiter.check("ip:success-test", session)
            await limiter.clear_for_key("ip:success-test", session)

        # A 9th successful login must still be allowed — counter stays at 0
        await limiter.check("ip:success-test", session)


@pytest.mark.asyncio
async def test_clear_for_key_resets_lockout(
    async_session_factory: async_sessionmaker[AsyncSession],
) -> None:
    """Successful login should reset the rate limit counter via clear_for_key."""
    limiter = DatabaseRateLimiter(max_attempts=8, window_seconds=60, type="password")

    async with async_session_factory() as session:
        # Record 8 failed attempts
        for _ in range(8):
            await limiter.check_and_record("ip:clear-test", session)

        # clear_for_key doesn't exist yet — this will raise AttributeError (xfail)
        await limiter.clear_for_key("ip:clear-test", session)

        # After clearing, should be able to attempt again
        await limiter.check_and_record("ip:clear-test", session)


@pytest.mark.asyncio
async def test_check_only_does_not_increment_counter(
    async_session_factory: async_sessionmaker[AsyncSession],
) -> None:
    """check() should only read, not write to the attempts table."""
    limiter = DatabaseRateLimiter(max_attempts=2, window_seconds=60, type="password")

    async with async_session_factory() as session:
        # check() without recording — 10 checks should NOT block
        for _ in range(10):
            await limiter.check("ip:check-only", session)

        # Still able to record
        await limiter.check_and_record("ip:check-only", session)


@pytest.mark.asyncio
async def test_record_failure_only_counts_failures(
    async_session_factory: async_sessionmaker[AsyncSession],
) -> None:
    """record_failure() should be the method called only when authentication fails."""
    limiter = DatabaseRateLimiter(max_attempts=3, window_seconds=60, type="password")

    async with async_session_factory() as session:
        # 2 failures recorded explicitly
        await limiter.record_failure("ip:failure-test", session)
        await limiter.record_failure("ip:failure-test", session)

        # 1 success (clear) in between
        await limiter.clear_for_key("ip:failure-test", session)

        # 1 more failure after success — should only be 1 total
        await limiter.record_failure("ip:failure-test", session)

        # Should NOT be blocked (only 1 failure since last success)
        await limiter.check("ip:failure-test", session)



from __future__ import annotations

import time
from collections.abc import Awaitable, Callable

import anyio

from app.core.config.settings import get_settings


class RateLimitHeadersCache:
    def __init__(self) -> None:
        self._cached_headers: dict[str, str] | None = None
        self._cached_at = 0.0
        self._lock = anyio.Lock()

    async def get(
        self,
        compute: Callable[[], Awaitable[dict[str, str]]],
    ) -> dict[str, str]:
        ttl = get_settings().usage_refresh_interval_seconds
        now = time.monotonic()
        if self._cached_headers is not None and now - self._cached_at < ttl:
            return self._cached_headers

        async with self._lock:
            now = time.monotonic()
            if self._cached_headers is not None and now - self._cached_at < ttl:
                return self._cached_headers

            headers = await compute()
            self._cached_headers = headers
            self._cached_at = now
            return headers

    async def invalidate(self) -> None:
        async with self._lock:
            self._cached_headers = None
            self._cached_at = 0.0


_rate_limit_headers_cache = RateLimitHeadersCache()


def get_rate_limit_headers_cache() -> RateLimitHeadersCache:
    return _rate_limit_headers_cache



"""add requested and actual service tiers to request_logs

Revision ID: 20260320_000000_add_request_log_requested_actual_tiers
Revises: 20260319_183000_normalize_sqlite_account_status_casing
Create Date: 2026-03-20
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260320_000000_add_request_log_requested_actual_tiers"
down_revision = "20260319_183000_normalize_sqlite_account_status_casing"
branch_labels = None
depends_on = None


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {column["name"] for column in inspector.get_columns(table_name)}


def upgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "request_logs")
    if not columns:
        return

    with op.batch_alter_table("request_logs") as batch_op:
        if "requested_service_tier" not in columns:
            batch_op.add_column(sa.Column("requested_service_tier", sa.String(), nullable=True))
        if "actual_service_tier" not in columns:
            batch_op.add_column(sa.Column("actual_service_tier", sa.String(), nullable=True))


def downgrade() -> None:
    bind = op.get_bind()
    columns = _columns(bind, "request_logs")
    if not columns:
        return

    with op.batch_alter_table("request_logs") as batch_op:
        if "actual_service_tier" in columns:
            batch_op.drop_column("actual_service_tier")
        if "requested_service_tier" in columns:
            batch_op.drop_column("requested_service_tier")



"""add api key assignment scope flag

Revision ID: 20260406_010000_add_api_key_assignment_scope_flag
Revises: 20260406_000000_add_api_key_account_assignments
Create Date: 2026-04-06
"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection

# revision identifiers, used by Alembic.
revision = "20260406_010000_add_api_key_assignment_scope_flag"
down_revision = "20260406_000000_add_api_key_account_assignments"
branch_labels = None
depends_on = None


def _table_exists(connection: Connection, table_name: str) -> bool:
    inspector = sa.inspect(connection)
    return inspector.has_table(table_name)


def _columns(connection: Connection, table_name: str) -> set[str]:
    inspector = sa.inspect(connection)
    if not inspector.has_table(table_name):
        return set()
    return {str(column["name"]) for column in inspector.get_columns(table_name) if column.get("name") is not None}


def upgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "api_keys"):
        return

    existing_columns = _columns(bind, "api_keys")
    with op.batch_alter_table("api_keys") as batch_op:
        if "account_assignment_scope_enabled" not in existing_columns:
            batch_op.add_column(
                sa.Column(
                    "account_assignment_scope_enabled",
                    sa.Boolean(),
                    nullable=False,
                    server_default=sa.false(),
                )
            )

    if _table_exists(bind, "api_key_accounts"):
        bind.execute(
            sa.text(
                """
                UPDATE api_keys
                SET account_assignment_scope_enabled = :enabled
                WHERE EXISTS (
                    SELECT 1
                    FROM api_key_accounts
                    WHERE api_key_accounts.api_key_id = api_keys.id
                )
                """
            ),
            {"enabled": True},
        )


def downgrade() -> None:
    bind = op.get_bind()
    if not _table_exists(bind, "api_keys"):
        return

    existing_columns = _columns(bind, "api_keys")
    with op.batch_alter_table("api_keys") as batch_op:
        if "account_assignment_scope_enabled" in existing_columns:
            batch_op.drop_column("account_assignment_scope_enabled")



from __future__ import annotations

import logging
from importlib import import_module

logger = logging.getLogger(__name__)

_otel_initialized = False


def init_tracing(service_name: str = "codex-lb", endpoint: str = "", app: object | None = None) -> bool:
    global _otel_initialized

    if _otel_initialized:
        return True

    try:
        trace = import_module("opentelemetry.trace")
        sdk_trace = import_module("opentelemetry.sdk.trace")
        sdk_resources = import_module("opentelemetry.sdk.resources")
        sdk_trace_export = import_module("opentelemetry.sdk.trace.export")
        Resource = getattr(sdk_resources, "Resource")
        SERVICE_NAME = getattr(sdk_resources, "SERVICE_NAME")
        TracerProvider = getattr(sdk_trace, "TracerProvider")
        BatchSpanProcessor = getattr(sdk_trace_export, "BatchSpanProcessor")

        provider = TracerProvider(resource=Resource.create({SERVICE_NAME: service_name}))

        if endpoint:
            try:
                exporter_module = import_module("opentelemetry.exporter.otlp.proto.grpc.trace_exporter")
                OTLPSpanExporter = getattr(exporter_module, "OTLPSpanExporter")

                exporter = OTLPSpanExporter(endpoint=endpoint)
                provider.add_span_processor(BatchSpanProcessor(exporter))
            except ImportError:
                logger.warning("OTLP exporter not available; tracing without export")

        trace.set_tracer_provider(provider)

        try:
            instrumentation_module = import_module("opentelemetry.instrumentation.fastapi")
            FastAPIInstrumentor = getattr(instrumentation_module, "FastAPIInstrumentor")

            if app is not None:
                FastAPIInstrumentor.instrument_app(app)
            else:
                FastAPIInstrumentor().instrument()
        except ImportError:
            pass
        except Exception:
            logger.exception("Failed to auto-instrument FastAPI")

        try:
            instrumentation_module = import_module("opentelemetry.instrumentation.aiohttp_client")
            AioHttpClientInstrumentor = getattr(instrumentation_module, "AioHttpClientInstrumentor")

            AioHttpClientInstrumentor().instrument()
        except ImportError:
            pass
        except Exception:
            logger.exception("Failed to auto-instrument aiohttp client")

        try:
            instrumentation_module = import_module("opentelemetry.instrumentation.sqlalchemy")
            SQLAlchemyInstrumentor = getattr(instrumentation_module, "SQLAlchemyInstrumentor")

            SQLAlchemyInstrumentor().instrument()
        except ImportError:
            pass
        except Exception:
            logger.exception("Failed to auto-instrument SQLAlchemy")

        _otel_initialized = True
        logger.info("OpenTelemetry tracing initialized (service=%s)", service_name)
        return True

    except ImportError:
        logger.warning(
            "opentelemetry packages not installed; tracing disabled. "
            "Install with: pip install codex-lb-cinamon[tracing]"
        )
        return False


def is_initialized() -> bool:
    return _otel_initialized


def get_current_trace_id() -> str | None:
    try:
        trace = import_module("opentelemetry.trace")

        span = getattr(trace, "get_current_span")()
        ctx = span.get_span_context()
        if ctx and ctx.is_valid:
            return format(ctx.trace_id, "032x")
    except Exception:
        logger.debug("Failed to get current trace ID", exc_info=True)
    return None


def get_current_span_id() -> str | None:
    try:
        trace = import_module("opentelemetry.trace")

        span = getattr(trace, "get_current_span")()
        ctx = span.get_span_context()
        if ctx and ctx.is_valid:
            return format(ctx.span_id, "016x")
    except Exception:
        logger.debug("Failed to get current span ID", exc_info=True)
    return None



"""Unit tests for LoadBalancer._select_with_stickiness cache-affinity fixes.

Covers:
- Fix 1+3: sticky session preservation when pinned account is temporarily down
- Fix 2: grace period for rate-limited accounts with imminent reset
"""

from __future__ import annotations

import time
from contextlib import asynccontextmanager
from typing import cast
from unittest.mock import AsyncMock

import pytest

from app.core.balancer import AccountState, RoutingStrategy
from app.db.models import Account, AccountStatus, StickySessionKind
from app.modules.proxy.load_balancer import LoadBalancer

pytestmark = pytest.mark.unit


def _active(account_id: str, used_percent: float = 10.0) -> AccountState:
    return AccountState(account_id, AccountStatus.ACTIVE, used_percent=used_percent)


def _rate_limited(
    account_id: str,
    reset_at: float | None = None,
    cooldown_until: float | None = None,
    used_percent: float | None = None,
) -> AccountState:
    return AccountState(
        account_id,
        AccountStatus.RATE_LIMITED,
        used_percent=used_percent,
        reset_at=reset_at,
        cooldown_until=cooldown_until,
        error_count=1,
        last_error_at=time.time(),
    )


def _make_sticky_repo(existing_account_id: str | None = None) -> AsyncMock:
    repo = AsyncMock()
    repo.get_account_id = AsyncMock(return_value=existing_account_id)
    repo.upsert = AsyncMock()
    repo.delete = AsyncMock()
    return repo


async def _invoke_stickiness(
    states: list[AccountState],
    sticky_key: str,
    sticky_repo: AsyncMock,
    *,
    sticky_kind: StickySessionKind = StickySessionKind.PROMPT_CACHE,
    reallocate_sticky: bool = False,
    sticky_max_age_seconds: int | None = 600,
    budget_threshold_pct: float = 95.0,
    routing_strategy: RoutingStrategy = "usage_weighted",
):
    """Wrapper that calls production LoadBalancer._select_with_stickiness.

    This delegates to the real implementation in LoadBalancer to avoid
    maintaining a replica that can drift from production.
    """

    @asynccontextmanager
    async def mock_repo_factory():
        yield AsyncMock()

    lb = LoadBalancer(mock_repo_factory)
    account_map = {s.account_id: cast(Account, AsyncMock()) for s in states}

    return await lb._select_with_stickiness(
        states=states,
        account_map=account_map,
        sticky_key=sticky_key,
        sticky_kind=sticky_kind,
        reallocate_sticky=reallocate_sticky,
        sticky_max_age_seconds=sticky_max_age_seconds,
        budget_threshold_pct=budget_threshold_pct,
        prefer_earlier_reset_accounts=False,
        routing_strategy=routing_strategy,
        sticky_repo=sticky_repo,
    )


# ---------------------------------------------------------------------------
# Fix 1+3: sticky session is preserved when pinned account is temporarily down
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_fallback_does_not_overwrite_sticky_when_pinned_rate_limited():
    """When the pinned account is rate-limited, a fallback is returned but the
    sticky mapping must NOT be overwritten — so the next request can return to
    the original account once it recovers."""
    now = time.time()
    acc_a = _rate_limited("a", cooldown_until=now + 60)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.upsert.assert_not_called()
    repo.delete.assert_not_called()


@pytest.mark.asyncio
async def test_all_accounts_unavailable_does_not_overwrite_sticky():
    """When the pinned account is down AND no fallback is available,
    the sticky mapping must still be preserved (not deleted or overwritten)."""
    now = time.time()
    acc_a = _rate_limited("a", cooldown_until=now + 60)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is None
    repo.upsert.assert_not_called()
    repo.delete.assert_not_called()


@pytest.mark.asyncio
async def test_fallback_overwrites_sticky_when_reallocate_sticky_true():
    """With reallocate_sticky=True (STICKY_THREAD), the sticky session IS
    deleted and the fallback IS persisted."""
    now = time.time()
    acc_a = _rate_limited("a", cooldown_until=now + 60)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        sticky_kind=StickySessionKind.STICKY_THREAD,
        reallocate_sticky=True,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once()
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.STICKY_THREAD)


@pytest.mark.asyncio
async def test_sticky_preserved_then_returns_to_original_on_recovery():
    """After a temporary fallback (without overwrite), the NEXT request
    returns to the original account once it recovers."""
    now = time.time()

    # Request 1: pinned account is rate-limited → fallback to b
    acc_a_down = _rate_limited("a", cooldown_until=now + 60)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    r1 = await _invoke_stickiness(
        [acc_a_down, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )
    assert r1.account is not None
    assert r1.account.account_id == "b"
    repo.upsert.assert_not_called()

    # Request 2: original account is back to active
    acc_a_up = _active("a", used_percent=5.0)
    acc_b2 = _active("b", used_percent=20.0)
    repo2 = _make_sticky_repo(existing_account_id="a")

    r2 = await _invoke_stickiness(
        [acc_a_up, acc_b2],
        "key1",
        repo2,
        reallocate_sticky=False,
    )
    assert r2.account is not None
    assert r2.account.account_id == "a"


@pytest.mark.asyncio
async def test_sticky_deleted_when_pinned_account_removed_from_pool():
    """When the pinned account is no longer in the account pool (deleted),
    the sticky session IS deleted and a new mapping IS persisted."""
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once_with("key1", kind=StickySessionKind.PROMPT_CACHE)
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)


# ---------------------------------------------------------------------------
# Pool exhaustion guard — prevent thrashing when all accounts are depleted
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_all_accounts_exhausted_keeps_pinned_no_thrashing():
    acc_a = _active("a", used_percent=96.0)
    acc_b = _active("b", used_percent=97.0)
    acc_c = _active("c", used_percent=98.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b, acc_c],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("key1", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_pool_exhausted_but_better_candidate_exists_reallocates():
    acc_a = _active("a", used_percent=96.0)
    acc_b = _active("b", used_percent=50.0)
    acc_c = _active("c", used_percent=97.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b, acc_c],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once_with("key1", kind=StickySessionKind.PROMPT_CACHE)
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_round_robin_pool_health_check_uses_round_robin_probe():
    now = time.time()
    acc_a = AccountState("a", AccountStatus.ACTIVE, used_percent=96.0, last_selected_at=now - 10)
    acc_b = AccountState("b", AccountStatus.ACTIVE, used_percent=50.0, last_selected_at=now - 1)
    acc_c = AccountState("c", AccountStatus.ACTIVE, used_percent=97.0, last_selected_at=now - 100)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b, acc_c],
        "key-round-robin",
        repo,
        reallocate_sticky=False,
        routing_strategy="round_robin",
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("key-round-robin", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_capacity_weighted_pool_health_check_uses_capacity_probe():
    acc_a = AccountState(
        "a",
        AccountStatus.ACTIVE,
        used_percent=96.0,
        secondary_used_percent=96.0,
        plan_type="pro",
        capacity_credits=50400.0,
    )
    acc_b = AccountState(
        "b",
        AccountStatus.ACTIVE,
        used_percent=50.0,
        secondary_used_percent=50.0,
        plan_type="free",
        capacity_credits=1134.0,
    )
    acc_c = AccountState(
        "c",
        AccountStatus.ACTIVE,
        used_percent=97.0,
        secondary_used_percent=97.0,
        plan_type="plus",
        capacity_credits=7560.0,
    )
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b, acc_c],
        "key-capacity-weighted",
        repo,
        reallocate_sticky=False,
        routing_strategy="capacity_weighted",
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("key-capacity-weighted", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_pool_exhausted_single_account_keeps_pinned():
    acc_a = _active("a", used_percent=96.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("key1", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_pool_exhausted_with_custom_threshold():
    acc_a = _active("a", used_percent=85.0)
    acc_b = _active("b", used_percent=88.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
        budget_threshold_pct=80.0,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("key1", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_pool_exhausted_candidate_with_none_usage_triggers_reallocation():
    acc_a = _active("a", used_percent=96.0)
    acc_b = AccountState("b", AccountStatus.ACTIVE, used_percent=None)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once_with("key1", kind=StickySessionKind.PROMPT_CACHE)
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_first_request_creates_sticky_mapping():
    """When no existing sticky mapping exists, the chosen account is persisted."""
    acc_a = _active("a", used_percent=10.0)
    acc_b = _active("b", used_percent=50.0)
    repo = _make_sticky_repo(existing_account_id=None)

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.upsert.assert_called_once_with("key1", "a", kind=StickySessionKind.PROMPT_CACHE)


# ---------------------------------------------------------------------------
# Fix 2: grace period for rate-limited pinned accounts with known reset_at
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_grace_period_returns_pinned_when_reset_imminent():
    """When the pinned account is RATE_LIMITED but resets within the grace
    window, it should be returned optimistically (not the fallback)."""
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 5)  # resets in 5s < grace 10s
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.upsert.assert_called_once_with("key1", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_grace_period_keeps_rate_limited_pinned_account_even_when_usage_is_100_pct():
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 5, used_percent=100.0)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("key1", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_grace_period_skipped_when_reset_far_away():
    """When the reset_at is well beyond the grace window, the grace period
    should NOT trigger — a fallback should be used instead."""
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 300)  # 5 minutes away
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.upsert.assert_not_called()


@pytest.mark.asyncio
async def test_grace_period_skipped_for_active_with_cooldown():
    """Accounts that apply_usage_quota resets to ACTIVE (with cooldown still
    active) must NOT be returned via the grace period — this avoids bypassing
    error-handling cooldowns."""
    now = time.time()
    # Simulates the apply_usage_quota path: status=ACTIVE, cooldown set
    acc_a = AccountState(
        "a",
        AccountStatus.ACTIVE,
        used_percent=10.0,
        cooldown_until=now + 2,
        error_count=1,
        last_error_at=now,
    )
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"


@pytest.mark.asyncio
async def test_grace_period_not_applied_for_reallocate_sticky():
    """Grace period only applies when reallocate_sticky=False."""
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 5)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        sticky_kind=StickySessionKind.STICKY_THREAD,
        reallocate_sticky=True,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once()


@pytest.mark.asyncio
async def test_grace_period_skipped_when_no_reset_at():
    """When the pinned account is RATE_LIMITED but has no reset_at (unknown
    recovery time), the grace period should not fire."""
    now = time.time()
    acc_a = _rate_limited("a", reset_at=None, cooldown_until=now + 2)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.upsert.assert_not_called()


# ---------------------------------------------------------------------------
# PR review issue 1: permanently down accounts must rebind sticky session
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_paused_pinned_account_persists_fallback():
    """PAUSED is permanent — the fallback MUST be persisted so the sticky
    session is rebound instead of pointing at a dead account forever."""
    acc_a = AccountState("a", AccountStatus.PAUSED)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_deactivated_pinned_account_persists_fallback():
    """DEACTIVATED is permanent — same rebind behaviour as PAUSED."""
    acc_a = AccountState("a", AccountStatus.DEACTIVATED, deactivation_reason="token expired")
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)


# ---------------------------------------------------------------------------
# PR review issue 2: grace period must not mutate original state
# ---------------------------------------------------------------------------


@pytest.mark.asyncio
async def test_grace_period_does_not_mutate_original_state():
    """The grace-period select_account call must operate on a copy so the
    original AccountState (synced to DB later) is not prematurely flipped
    to ACTIVE."""
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 5)  # within grace window
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    original_status = acc_a.status
    original_reset_at = acc_a.reset_at

    await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert acc_a.status == original_status
    assert acc_a.reset_at == original_reset_at


@pytest.mark.asyncio
async def test_codex_session_persists_fallback_during_outage():
    """CODEX_SESSION is durable (no TTL). When the pinned account is
    temporarily down, the fallback MUST be persisted so the session
    sticks to one account instead of bouncing across random fallbacks."""
    now = time.time()
    acc_a = _rate_limited("a", cooldown_until=now + 60)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "session_123",
        repo,
        sticky_kind=StickySessionKind.CODEX_SESSION,
        reallocate_sticky=False,
        sticky_max_age_seconds=None,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.upsert.assert_called_once_with("session_123", "b", kind=StickySessionKind.CODEX_SESSION)


@pytest.mark.asyncio
async def test_rate_limit_far_away_does_not_reallocate_codex_session_affinity():
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 1200)
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "session_123",
        repo,
        sticky_kind=StickySessionKind.CODEX_SESSION,
        reallocate_sticky=False,
        sticky_max_age_seconds=None,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("session_123", "b", kind=StickySessionKind.CODEX_SESSION)


@pytest.mark.asyncio
async def test_rate_limit_far_away_without_fallback_preserves_codex_session_affinity():
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 1200)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a],
        "session_123",
        repo,
        sticky_kind=StickySessionKind.CODEX_SESSION,
        reallocate_sticky=False,
        sticky_max_age_seconds=None,
    )

    assert result.account is None
    repo.delete.assert_not_called()
    repo.upsert.assert_not_called()


@pytest.mark.asyncio
async def test_budget_exhaustion_triggers_reallocation():
    """When the pinned account has < 5% budget remaining (used_percent >= 95%),
    it should be reallocated to a different account."""
    acc_a = _active("a", used_percent=96.0)  # 96% used = < 4% remaining
    acc_b = _active("b", used_percent=50.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once_with("key1", kind=StickySessionKind.PROMPT_CACHE)
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_budget_threshold_80_triggers_at_85_percent():
    acc_a = _active("a", used_percent=85.0)
    acc_b = _active("b", used_percent=50.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
        budget_threshold_pct=80.0,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once_with("key1", kind=StickySessionKind.PROMPT_CACHE)
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_budget_threshold_95_no_reallocation_at_85_percent():
    acc_a = _active("a", used_percent=85.0)
    acc_b = _active("b", used_percent=50.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
        budget_threshold_pct=95.0,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_called_once_with("key1", "a", kind=StickySessionKind.PROMPT_CACHE)


@pytest.mark.asyncio
async def test_budget_threshold_does_not_reallocate_codex_session_affinity():
    acc_a = _active("a", used_percent=96.0)
    acc_b = _active("b", used_percent=50.0)
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "codex-session-123",
        repo,
        sticky_kind=StickySessionKind.CODEX_SESSION,
        reallocate_sticky=False,
        sticky_max_age_seconds=None,
        budget_threshold_pct=95.0,
    )

    assert result.account is not None
    assert result.account.account_id == "a"
    repo.delete.assert_not_called()
    repo.upsert.assert_not_called()


@pytest.mark.asyncio
async def test_rate_limit_far_away_triggers_reallocation():
    """When the pinned account's rate limit reset is more than 10 minutes away,
    it should be reallocated to avoid waiting."""
    now = time.time()
    acc_a = _rate_limited("a", reset_at=now + 900)  # resets in 15 minutes > 10 min threshold
    acc_b = _active("b")
    repo = _make_sticky_repo(existing_account_id="a")

    result = await _invoke_stickiness(
        [acc_a, acc_b],
        "key1",
        repo,
        reallocate_sticky=False,
    )

    assert result.account is not None
    assert result.account.account_id == "b"
    repo.delete.assert_called_once_with("key1", kind=StickySessionKind.PROMPT_CACHE)
    repo.upsert.assert_called_once_with("key1", "b", kind=StickySessionKind.PROMPT_CACHE)



from __future__ import annotations

import logging
from datetime import datetime
from typing import Protocol

from app.core.auth import DEFAULT_PLAN, OpenAIAuthClaims, extract_id_token_claims
from app.core.auth.refresh import RefreshError, refresh_access_token, should_refresh
from app.core.balancer import PERMANENT_FAILURE_CODES
from app.core.crypto import TokenEncryptor
from app.core.plan_types import coerce_account_plan_type
from app.core.utils.time import utcnow
from app.db.models import Account, AccountStatus


class AccountsRepositoryPort(Protocol):
    async def get_by_id(self, account_id: str) -> Account | None: ...

    async def update_status(
        self,
        account_id: str,
        status: AccountStatus,
        deactivation_reason: str | None = None,
    ) -> bool: ...

    async def update_tokens(
        self,
        account_id: str,
        access_token_encrypted: bytes,
        refresh_token_encrypted: bytes,
        id_token_encrypted: bytes,
        last_refresh: datetime,
        plan_type: str | None = None,
        email: str | None = None,
        chatgpt_account_id: str | None = None,
    ) -> bool: ...


logger = logging.getLogger(__name__)


class AuthManager:
    def __init__(self, repo: AccountsRepositoryPort) -> None:
        self._repo = repo
        self._encryptor = TokenEncryptor()

    async def ensure_fresh(self, account: Account, *, force: bool = False) -> Account:
        if force or should_refresh(account.last_refresh):
            account = await self.refresh_account(account)
        return await self._ensure_chatgpt_account_id(account)

    async def refresh_account(self, account: Account) -> Account:
        refresh_token = self._encryptor.decrypt(account.refresh_token_encrypted)
        try:
            result = await refresh_access_token(refresh_token)
        except RefreshError as exc:
            if exc.is_permanent:
                reason = PERMANENT_FAILURE_CODES.get(exc.code, exc.message)
                await self._repo.update_status(account.id, AccountStatus.DEACTIVATED, reason)
                account.status = AccountStatus.DEACTIVATED
                account.deactivation_reason = reason
            raise

        account.access_token_encrypted = self._encryptor.encrypt(result.access_token)
        account.refresh_token_encrypted = self._encryptor.encrypt(result.refresh_token)
        account.id_token_encrypted = self._encryptor.encrypt(result.id_token)
        account.last_refresh = utcnow()
        if result.account_id:
            account.chatgpt_account_id = result.account_id
        if result.plan_type is not None:
            account.plan_type = coerce_account_plan_type(
                result.plan_type,
                account.plan_type or DEFAULT_PLAN,
            )
        elif not account.plan_type:
            account.plan_type = DEFAULT_PLAN
        if result.email:
            account.email = result.email

        await self._repo.update_tokens(
            account.id,
            access_token_encrypted=account.access_token_encrypted,
            refresh_token_encrypted=account.refresh_token_encrypted,
            id_token_encrypted=account.id_token_encrypted,
            last_refresh=account.last_refresh,
            plan_type=account.plan_type,
            email=account.email,
            chatgpt_account_id=account.chatgpt_account_id,
        )
        return account

    async def _ensure_chatgpt_account_id(self, account: Account) -> Account:
        if account.chatgpt_account_id:
            return account
        try:
            id_token = self._encryptor.decrypt(account.id_token_encrypted)
        except Exception:
            return account
        raw_account_id = _chatgpt_account_id_from_id_token(id_token)
        if not raw_account_id:
            return account

        account.chatgpt_account_id = raw_account_id
        try:
            await self._repo.update_tokens(
                account.id,
                access_token_encrypted=account.access_token_encrypted,
                refresh_token_encrypted=account.refresh_token_encrypted,
                id_token_encrypted=account.id_token_encrypted,
                last_refresh=account.last_refresh,
                plan_type=account.plan_type,
                email=account.email,
                chatgpt_account_id=raw_account_id,
            )
        except Exception:
            logger.warning("Failed to persist chatgpt_account_id account_id=%s", account.id, exc_info=True)
        return account


def _chatgpt_account_id_from_id_token(id_token: str) -> str | None:
    claims = extract_id_token_claims(id_token)
    auth_claims = claims.auth or OpenAIAuthClaims()
    return auth_claims.chatgpt_account_id or claims.chatgpt_account_id



from __future__ import annotations

import json
from collections.abc import AsyncIterator, Iterator
from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine

from app.db.models import AdditionalUsageHistory, Base
from app.modules.usage.additional_quota_keys import clear_additional_quota_registry_cache
from app.modules.usage.repository import AdditionalUsageRepository

pytestmark = pytest.mark.unit


@pytest.fixture
async def async_session() -> AsyncIterator[AsyncSession]:
    engine = create_async_engine("sqlite+aiosqlite:///:memory:")
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    async_session_factory = async_sessionmaker(engine, expire_on_commit=False)
    session = async_session_factory()
    try:
        yield session
    finally:
        await session.close()

    await engine.dispose()


@pytest.fixture(autouse=True)
def _clear_registry_cache() -> Iterator[None]:
    clear_additional_quota_registry_cache()
    yield
    clear_additional_quota_registry_cache()


def _write_registry(path: Path, *, quota_key: str, quota_key_aliases: list[str] | None = None) -> None:
    path.write_text(
        json.dumps(
            [
                {
                    "quota_key": quota_key,
                    "quota_key_aliases": quota_key_aliases or [],
                    "display_label": "Spark Enterprise",
                    "model_ids": ["gpt-5.3-codex-spark"],
                    "limit_name_aliases": ["codex_other"],
                    "metered_feature_aliases": ["codex_bengalfox"],
                }
            ]
        ),
        encoding="utf-8",
    )


@pytest.mark.asyncio
async def test_add_entry(async_session: AsyncSession) -> None:
    """Test adding an entry to additional usage history."""
    repo = AdditionalUsageRepository(async_session)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=50.0,
        reset_at=1735689600,
        window_minutes=1,
    )

    stmt = select(AdditionalUsageHistory).where(AdditionalUsageHistory.account_id == "acc_1")
    result = await async_session.execute(stmt)
    entries = result.scalars().all()
    assert len(entries) == 1
    assert entries[0].account_id == "acc_1"
    assert entries[0].limit_name == "requests_per_minute"
    assert entries[0].used_percent == 50.0


@pytest.mark.asyncio
async def test_latest_by_account_returns_most_recent_per_account(async_session: AsyncSession) -> None:
    """Test that latest_by_account returns only the most recent entry per account."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)
    old_time = now - timedelta(hours=1)

    # Add multiple entries for same account, different times
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=old_time,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=50.0,
        recorded_at=now,
    )

    result = await repo.latest_by_account(limit_name="requests_per_minute", window="1m")

    assert len(result) == 1
    assert "acc_1" in result
    assert result["acc_1"].used_percent == 50.0


@pytest.mark.asyncio
async def test_latest_by_account_multiple_accounts(async_session: AsyncSession) -> None:
    """Test latest_by_account with multiple accounts."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_2",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=60.0,
        recorded_at=now,
    )

    result = await repo.latest_by_account(limit_name="requests_per_minute", window="1m")

    assert len(result) == 2
    assert result["acc_1"].used_percent == 30.0
    assert result["acc_2"].used_percent == 60.0


@pytest.mark.asyncio
async def test_latest_by_account_empty_when_no_data(async_session: AsyncSession) -> None:
    """Test latest_by_account returns empty dict when no data exists."""
    repo = AdditionalUsageRepository(async_session)

    result = await repo.latest_by_account(limit_name="requests_per_minute", window="1m")

    assert result == {}


@pytest.mark.asyncio
async def test_latest_by_account_filters_by_limit_name(async_session: AsyncSession) -> None:
    """Test that latest_by_account only returns requested limit_name."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="1h",
        used_percent=60.0,
        recorded_at=now,
    )

    result = await repo.latest_by_account(limit_name="requests_per_minute", window="1m")

    assert len(result) == 1
    assert result["acc_1"].limit_name == "requests_per_minute"
    assert result["acc_1"].used_percent == 30.0


@pytest.mark.asyncio
async def test_latest_by_account_filters_by_window(async_session: AsyncSession) -> None:
    """Test that latest_by_account filters by window."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="5m",
        used_percent=60.0,
        recorded_at=now,
    )

    result = await repo.latest_by_account(limit_name="requests_per_minute", window="1m")

    assert len(result) == 1
    assert result["acc_1"].window == "1m"
    assert result["acc_1"].used_percent == 30.0


@pytest.mark.asyncio
async def test_latest_by_account_canonicalizes_legacy_limit_name_alias(async_session: AsyncSession) -> None:
    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="codex_other",
        metered_feature="codex_bengalfox",
        window="primary",
        used_percent=30.0,
        recorded_at=now,
    )

    result = await repo.latest_by_account(limit_name="GPT-5.3-Codex-Spark", window="primary")

    assert list(result) == ["acc_1"]
    assert result["acc_1"].quota_key == "codex_spark"
    assert result["acc_1"].limit_name == "codex_other"


@pytest.mark.asyncio
async def test_latest_by_account_reads_rows_under_legacy_quota_key_alias(
    async_session: AsyncSession,
    monkeypatch,
    tmp_path: Path,
) -> None:
    registry = tmp_path / "additional_quota_registry.json"
    _write_registry(registry, quota_key="spark_enterprise", quota_key_aliases=["codex_spark"])
    monkeypatch.setenv("CODEX_LB_ADDITIONAL_QUOTA_REGISTRY_FILE", str(registry))
    clear_additional_quota_registry_cache()

    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    async_session.add(
        AdditionalUsageHistory(
            account_id="acc_legacy",
            quota_key="codex_spark",
            limit_name="codex_other",
            metered_feature="codex_bengalfox",
            window="primary",
            used_percent=42.0,
            recorded_at=now,
        )
    )
    await async_session.commit()

    result = await repo.latest_by_account(quota_key="spark_enterprise", window="primary")

    assert list(result) == ["acc_legacy"]
    assert result["acc_legacy"].quota_key == "codex_spark"
    assert await repo.list_quota_keys() == ["spark_enterprise"]


@pytest.mark.asyncio
async def test_list_limit_names_returns_distinct_names(async_session: AsyncSession) -> None:
    """Test list_limit_names returns distinct limit names."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_2",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=60.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="1h",
        used_percent=40.0,
        recorded_at=now,
    )

    result = await repo.list_limit_names()

    assert len(result) == 2
    assert "requests_per_minute" in result
    assert "requests_per_hour" in result


@pytest.mark.asyncio
async def test_list_limit_names_empty_when_no_data(async_session: AsyncSession) -> None:
    """Test list_limit_names returns empty list when no data exists."""
    repo = AdditionalUsageRepository(async_session)

    result = await repo.list_limit_names()

    assert result == []


@pytest.mark.asyncio
async def test_history_since_returns_time_series(async_session: AsyncSession) -> None:
    """Test history_since returns time-series entries in order."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)
    t1 = now - timedelta(hours=2)
    t2 = now - timedelta(hours=1)
    t3 = now

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=t1,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=40.0,
        recorded_at=t2,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=50.0,
        recorded_at=t3,
    )

    since = now - timedelta(hours=3)
    result = await repo.history_since(
        account_id="acc_1",
        limit_name="requests_per_minute",
        window="1m",
        since=since,
    )

    assert len(result) == 3
    assert result[0].used_percent == 30.0
    assert result[1].used_percent == 40.0
    assert result[2].used_percent == 50.0


@pytest.mark.asyncio
async def test_history_since_filters_by_since_time(async_session: AsyncSession) -> None:
    """Test history_since only returns entries after since time."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)
    t1 = now - timedelta(hours=2)
    t2 = now - timedelta(hours=1)
    t3 = now

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=t1,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=40.0,
        recorded_at=t2,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=50.0,
        recorded_at=t3,
    )

    since = now - timedelta(hours=1, minutes=30)
    result = await repo.history_since(
        account_id="acc_1",
        limit_name="requests_per_minute",
        window="1m",
        since=since,
    )

    assert len(result) == 2
    assert result[0].used_percent == 40.0
    assert result[1].used_percent == 50.0


@pytest.mark.asyncio
async def test_history_since_filters_by_account_id(async_session: AsyncSession) -> None:
    """Test history_since only returns entries for specified account."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_2",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=60.0,
        recorded_at=now,
    )

    since = now - timedelta(hours=1)
    result = await repo.history_since(
        account_id="acc_1",
        limit_name="requests_per_minute",
        window="1m",
        since=since,
    )

    assert len(result) == 1
    assert result[0].account_id == "acc_1"
    assert result[0].used_percent == 30.0


@pytest.mark.asyncio
async def test_history_since_filters_by_limit_name(async_session: AsyncSession) -> None:
    """Test history_since only returns entries for specified limit_name."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="1h",
        used_percent=60.0,
        recorded_at=now,
    )

    since = now - timedelta(hours=1)
    result = await repo.history_since(
        account_id="acc_1",
        limit_name="requests_per_minute",
        window="1m",
        since=since,
    )

    assert len(result) == 1
    assert result[0].limit_name == "requests_per_minute"


@pytest.mark.asyncio
async def test_history_since_filters_by_window(async_session: AsyncSession) -> None:
    """Test history_since only returns entries for specified window."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="5m",
        used_percent=60.0,
        recorded_at=now,
    )

    since = now - timedelta(hours=1)
    result = await repo.history_since(
        account_id="acc_1",
        limit_name="requests_per_minute",
        window="1m",
        since=since,
    )

    assert len(result) == 1
    assert result[0].window == "1m"


@pytest.mark.asyncio
async def test_history_since_empty_when_no_data(async_session: AsyncSession) -> None:
    """Test history_since returns empty list when no data exists."""
    repo = AdditionalUsageRepository(async_session)

    now = datetime.now(tz=timezone.utc)
    since = now - timedelta(hours=1)

    result = await repo.history_since(
        account_id="acc_1",
        limit_name="requests_per_minute",
        window="1m",
        since=since,
    )

    assert result == []


@pytest.mark.asyncio
async def test_history_since_canonicalizes_legacy_limit_name_alias(async_session: AsyncSession) -> None:
    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_1",
        limit_name="codex_other",
        metered_feature="codex_bengalfox",
        window="primary",
        used_percent=30.0,
        recorded_at=now - timedelta(minutes=5),
    )
    await repo.add_entry(
        account_id="acc_1",
        limit_name="codex_other",
        metered_feature="codex_bengalfox",
        window="primary",
        used_percent=45.0,
        recorded_at=now,
    )

    result = await repo.history_since(
        account_id="acc_1",
        limit_name="GPT-5.3-Codex-Spark",
        window="primary",
        since=now - timedelta(hours=1),
    )

    assert [entry.used_percent for entry in result] == [30.0, 45.0]
    assert all(entry.quota_key == "codex_spark" for entry in result)


@pytest.mark.asyncio
async def test_history_since_reads_rows_under_legacy_quota_key_alias(
    async_session: AsyncSession,
    monkeypatch,
    tmp_path: Path,
) -> None:
    registry = tmp_path / "additional_quota_registry.json"
    _write_registry(registry, quota_key="spark_enterprise", quota_key_aliases=["codex_spark"])
    monkeypatch.setenv("CODEX_LB_ADDITIONAL_QUOTA_REGISTRY_FILE", str(registry))
    clear_additional_quota_registry_cache()

    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    async_session.add_all(
        [
            AdditionalUsageHistory(
                account_id="acc_legacy",
                quota_key="codex_spark",
                limit_name="codex_other",
                metered_feature="codex_bengalfox",
                window="primary",
                used_percent=20.0,
                recorded_at=now - timedelta(minutes=5),
            ),
            AdditionalUsageHistory(
                account_id="acc_legacy",
                quota_key="codex_spark",
                limit_name="codex_other",
                metered_feature="codex_bengalfox",
                window="primary",
                used_percent=45.0,
                recorded_at=now,
            ),
        ]
    )
    await async_session.commit()

    result = await repo.history_since(
        account_id="acc_legacy",
        quota_key="spark_enterprise",
        window="primary",
        since=now - timedelta(hours=1),
    )

    assert [entry.used_percent for entry in result] == [20.0, 45.0]
    assert all(entry.quota_key == "codex_spark" for entry in result)


@pytest.mark.asyncio
async def test_delete_for_account_removes_all_rows(async_session: AsyncSession) -> None:
    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="1h",
        used_percent=60.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_keep",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=45.0,
        recorded_at=now,
    )

    await repo.delete_for_account("acc_delete")

    result = await async_session.execute(select(AdditionalUsageHistory).order_by(AdditionalUsageHistory.account_id))
    entries = result.scalars().all()
    assert len(entries) == 1
    assert entries[0].account_id == "acc_keep"


@pytest.mark.asyncio
async def test_delete_for_account_and_limit_removes_only_matching_rows(async_session: AsyncSession) -> None:
    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="5m",
        used_percent=35.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="1h",
        used_percent=60.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_keep",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="1m",
        used_percent=45.0,
        recorded_at=now,
    )

    await repo.delete_for_account_and_limit("acc_delete", "requests_per_minute")

    result = await async_session.execute(
        select(AdditionalUsageHistory).order_by(
            AdditionalUsageHistory.account_id,
            AdditionalUsageHistory.limit_name,
            AdditionalUsageHistory.window,
        )
    )
    entries = result.scalars().all()

    assert len(entries) == 2
    assert [(entry.account_id, entry.limit_name, entry.window) for entry in entries] == [
        ("acc_delete", "requests_per_hour", "1h"),
        ("acc_keep", "requests_per_minute", "1m"),
    ]


@pytest.mark.asyncio
async def test_delete_for_account_and_limit_canonicalizes_legacy_alias(async_session: AsyncSession) -> None:
    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_delete",
        limit_name="codex_other",
        metered_feature="codex_bengalfox",
        window="primary",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="primary",
        used_percent=60.0,
        recorded_at=now,
    )

    await repo.delete_for_account_and_limit("acc_delete", "GPT-5.3-Codex-Spark")

    result = await async_session.execute(
        select(AdditionalUsageHistory).order_by(
            AdditionalUsageHistory.account_id,
            AdditionalUsageHistory.limit_name,
            AdditionalUsageHistory.window,
        )
    )
    entries = result.scalars().all()

    assert [(entry.account_id, entry.limit_name, entry.window) for entry in entries] == [
        ("acc_delete", "requests_per_hour", "primary"),
    ]


@pytest.mark.asyncio
async def test_delete_for_account_and_quota_key_removes_rows_under_legacy_quota_key_alias(
    async_session: AsyncSession,
    monkeypatch,
    tmp_path: Path,
) -> None:
    registry = tmp_path / "additional_quota_registry.json"
    _write_registry(registry, quota_key="spark_enterprise", quota_key_aliases=["codex_spark"])
    monkeypatch.setenv("CODEX_LB_ADDITIONAL_QUOTA_REGISTRY_FILE", str(registry))
    clear_additional_quota_registry_cache()

    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    async_session.add(
        AdditionalUsageHistory(
            account_id="acc_delete",
            quota_key="codex_spark",
            limit_name="codex_other",
            metered_feature="codex_bengalfox",
            window="primary",
            used_percent=30.0,
            recorded_at=now,
        )
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="primary",
        used_percent=60.0,
        recorded_at=now,
    )
    await async_session.commit()

    await repo.delete_for_account_and_quota_key("acc_delete", "spark_enterprise")

    result = await async_session.execute(
        select(AdditionalUsageHistory).order_by(
            AdditionalUsageHistory.account_id,
            AdditionalUsageHistory.limit_name,
            AdditionalUsageHistory.window,
        )
    )
    entries = result.scalars().all()

    assert [(entry.account_id, entry.limit_name, entry.window) for entry in entries] == [
        ("acc_delete", "requests_per_hour", "primary"),
    ]


@pytest.mark.asyncio
async def test_delete_for_account_limit_window_removes_only_matching_window(async_session: AsyncSession) -> None:
    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="primary",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_minute",
        metered_feature="api_calls",
        window="secondary",
        used_percent=35.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="requests_per_hour",
        metered_feature="api_calls",
        window="primary",
        used_percent=60.0,
        recorded_at=now,
    )

    await repo.delete_for_account_limit_window("acc_delete", "requests_per_minute", "secondary")

    result = await async_session.execute(
        select(AdditionalUsageHistory).order_by(
            AdditionalUsageHistory.account_id,
            AdditionalUsageHistory.limit_name,
            AdditionalUsageHistory.window,
        )
    )
    entries = result.scalars().all()

    assert [(entry.account_id, entry.limit_name, entry.window) for entry in entries] == [
        ("acc_delete", "requests_per_hour", "primary"),
        ("acc_delete", "requests_per_minute", "primary"),
    ]


@pytest.mark.asyncio
async def test_delete_for_account_limit_window_canonicalizes_legacy_alias(async_session: AsyncSession) -> None:
    repo = AdditionalUsageRepository(async_session)
    now = datetime.now(tz=timezone.utc)

    await repo.add_entry(
        account_id="acc_delete",
        limit_name="codex_other",
        metered_feature="codex_bengalfox",
        window="primary",
        used_percent=30.0,
        recorded_at=now,
    )
    await repo.add_entry(
        account_id="acc_delete",
        limit_name="codex_other",
        metered_feature="codex_bengalfox",
        window="secondary",
        used_percent=35.0,
        recorded_at=now,
    )

    await repo.delete_for_account_limit_window("acc_delete", "GPT-5.3-Codex-Spark", "secondary")

    result = await async_session.execute(
        select(AdditionalUsageHistory).order_by(
            AdditionalUsageHistory.account_id,
            AdditionalUsageHistory.limit_name,
            AdditionalUsageHistory.window,
        )
    )
    entries = result.scalars().all()

    assert [(entry.account_id, entry.limit_name, entry.window) for entry in entries] == [
        ("acc_delete", "codex_other", "primary"),
    ]
